package lsedit;

import java.util.Enumeration;
import java.util.Vector;

import java.io.BufferedWriter;
import java.io.IOException;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.Rectangle;
import java.awt.Toolkit;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JLabel;

public class MatrixLayout extends LandscapeLayouter  implements ToolBarEventHandler {

	static protected final int BORDER           = 0;	
	static protected final int XGAP             = 1;
	static protected final int YGAP             = 2;

	protected final static String[] m_textfield_tags = 
							{
								"matrix:border=",
								"matrix:xgap=",
								"matrix:ygap="
							};

	protected final static String[] m_textfield_titles = 
							{
								"Border fraction of total:",
								"Horizontal whitespace fraction of grid:",
								"Vertical whitespace fraction of grid:"
							};					

	protected final static String[] m_textfield_resets = 
							{
								"0.0333",
								"0.2",
								"0.2"
							};

	protected static String[] m_textfield_defaults = 
							{
								"0.0333",
								"0.2",
								"0.2"
							};

	static	JTextField[]	m_textfields;

	static protected final int BUTTON_OK      = 0;
	static protected final int BUTTON_CANCEL  = 1;
	static protected final int BUTTON_HELP    = 2;
	static protected final int BUTTON_DEFAULT = 3;
	static protected final int BUTTON_SET     = 4;
	static protected final int BUTTON_RESET   = 5;

	protected final static String[] m_button_titles =
							{
								"Ok",
								"Cancel",
								"Help",
								"Default",
								"Set",
								"Reset"
							};

	protected final static String[] m_button_tips =
							{
								null,
								null,
								null,
								"Use remembered default",
								"Set default to current",
								"Set default to initial"
							};

	protected static double parameterDouble(int i)
	{
		String	string;
		double	dval;

		if (m_textfields == null) {
			string = m_textfield_defaults[i];
		} else {
			string = m_textfields[i].getText();
		}
		string = string.trim();

		try {
			dval = Double.parseDouble(string);
		} catch (Throwable exception) {
			dval = 0;
		}
		return dval;
	}

	public static double getBorder()
	{
		return parameterDouble(BORDER);
	}

	public static double getXGap()
	{
		return parameterDouble(XGAP);
	}

	public static double getYGap()
	{
		return parameterDouble(YGAP);
	}

	public String getTag()
	{
		return "matrix:";
	}

	public void load(String string)
	{
		String	tag;
		int		i;

		for (i = 0; i < m_textfield_tags.length; ++i) {
			tag = m_textfield_tags[i];
			if (string.startsWith(tag)) {
				m_textfield_defaults[i] = string.substring(tag.length());
				return;
		}	}
	}

	public void save(BufferedWriter bufferedWriter) throws IOException
	{
		String	string;
		int		i;

		for (i = 0; i < m_textfield_tags.length; ++i) {
			string = m_textfield_tags[i];
			bufferedWriter.write(string, 0, string.length());
			string = m_textfield_defaults[i];
			bufferedWriter.write(string, 0, string.length());
			bufferedWriter.newLine();
	}	}

	class MatrixConfigure extends JDialog implements ActionListener {

		class LayoutImage extends JComponent implements MouseListener
		{
			Rectangle[] m_boxs;
			int			m_rows = 2;
			boolean		m_validated;


			public LayoutImage()
			{
				super();

				Dimension	dimension = new Dimension(400,240);
				int			i;

				setLayout(null);
				setPreferredSize(dimension);
				setMinimumSize(dimension);
				setMaximumSize(dimension);
				setSize(dimension);

				m_boxs      = new Rectangle[25];
				m_validated = false;
				addMouseListener(this);
				setVisible(true);
			}
			
			public void validate()
			{
				Rectangle[]	boxs       = m_boxs;
				int			cnt        = boxs.length;
				int			dimension  = m_rows;;				
				int			width      = getWidth();
				int			height     = getHeight();
				double		new_border = MatrixLayout.getBorder();
				double		new_xgap   = getXGap();
				double		new_ygap   = getYGap();

				int			x, y, w, h, gapw, gaph, w1, h1, row, col, i, gaps;
				double		dgaps;
				Rectangle	rectangle;

				x      = (int) (((double) width)  * new_border / 2.0);
				y      = (int) (((double) height) * new_border / 2.0);
				w      = width  - 2 * x;
				h      = height - 2 * y;

				m_validated = false;

				if (w < 1 || h < 1) {
					return;
				}

				gaps      = dimension - 1;
				dgaps     = (double) gaps;
				gapw      = (int) ((new_xgap * ((double) w))/dgaps); // Size of the gaps
				gaph      = (int) ((new_ygap * ((double) h))/dgaps);
				w1        = (w - (gapw * gaps)) / dimension;
				h1        = (h - (gaph * gaps)) / dimension;
					
				if (w1 < 1 || h1 < 1) {
					return;
				}

				row = col = 0;
				cnt = dimension * dimension;
				for (i = 0; i < cnt; ++i) {
					rectangle = (Rectangle) boxs[i];
					if (rectangle == null) {
						boxs[i] = rectangle = new Rectangle();
					}
					rectangle.x = x;
					if (col != 0) {
						rectangle.x += col*(w1 + gapw);
					}
					rectangle.y = y;
					if (row != 0) {
						rectangle.y += row*(h1 + gaph);
					}
					rectangle.width  = w1;
					rectangle.height = h1;
					if (++col == dimension) {
						++row;
						col = 0;
				}	}

				m_validated = true;
			}

			
			public void paintComponent(Graphics g)
			{
				int	dimension, w, h, row, col, cnt;

				w = getWidth();
				h = getHeight();

				g.setColor(Color.black);
				g.drawRect(0, 0, w-1, h-1);

				if (m_validated) {
					Rectangle[]	boxs = m_boxs;
					Rectangle	rectangle;
					int			i;

					g.setColor(Color.blue);
					dimension = m_rows;
					row = col = 0;
					cnt = dimension * dimension;
					for (i = 0; i < cnt; ++i) {
						rectangle = boxs[i];
						g.drawRect(rectangle.x, rectangle.y, rectangle.width, rectangle.height); 
						if (++col == dimension) {
							++row;
							col = 0;
					}	}
					if (dimension == 2) {
						String		s     = "Click to change dimensions";
						FontMetrics fm    = g.getFontMetrics();
						int			width = fm.stringWidth(s);

						g.setColor(Color.black);
						g.drawString(s, (w - width)/2, h/2);
				}	}
			}

			// MouseListener interface

			public void mouseClicked(MouseEvent ev)
			{
			}

			public void mouseEntered(MouseEvent ev)
			{
			}

			public void mouseExited(MouseEvent ev)
			{
			}

			public void mousePressed(MouseEvent ev)
			{
				if (ev.isMetaDown()) {
					--m_rows;
					if (m_rows < 2) {
						m_rows = 5;
					}
				} else {
					++m_rows;
					if (m_rows > 5) {
						m_rows = 2;
				}	}
				validate();
				repaint();
			}

			public void mouseReleased(MouseEvent ev)
			{
			}
		}

		protected LayoutImage		m_layoutImage;
		protected JLabel			m_message;
		protected JButton[]			m_buttons;
		boolean	  m_ret;

		public MatrixConfigure()
		{
			super(getLs().getFrame(), "Matrix Whitespace", true);

			Container			contentPane;
			Font				font, bold;
			JLabel				label;
			JTextField			textfield;
			JButton				button;
			String				tip;
			int					i;

			m_ret    = false;

			font   = FontCache.getDialogFont();
			bold   = font.deriveFont(Font.BOLD);

			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(font);

			JPanel topPanel    = new JPanel();
			JPanel labelPanel  = new JPanel();
			JPanel valuePanel  = new JPanel();

			GridLayout gridLayout;

			topPanel.setLayout( new BorderLayout() );
			gridLayout = new GridLayout(3, 1, 0, 10);
			labelPanel.setLayout(gridLayout);

			gridLayout = new GridLayout(3,1, 0, 10);
			valuePanel.setLayout(gridLayout);


			if (m_textfields == null) {
				m_textfields = new JTextField[m_textfield_tags.length];
				for (i = 0; i < m_textfield_tags.length; ++i) {
					m_textfields[i] = new JTextField(m_textfield_defaults[i],  20);
			}	}

			for (i = 0; i < m_textfield_tags.length; ++i) {

				label = new JLabel(m_textfield_titles[i], JLabel.RIGHT);
				label.setFont(bold);
				labelPanel.add(label);
				textfield = m_textfields[i];
				textfield.addActionListener(this);
				textfield.setFont(font);
				valuePanel.add(textfield);
			}

			topPanel.add( BorderLayout.WEST,   labelPanel);
			topPanel.add( BorderLayout.EAST,   valuePanel);

			contentPane = getContentPane();
			
			contentPane.add( BorderLayout.NORTH, topPanel );

			JPanel centerPanel = new JPanel();
			centerPanel.setLayout( new BorderLayout() );


			m_layoutImage = new LayoutImage();
			centerPanel.add( BorderLayout.NORTH, m_layoutImage);
			m_layoutImage.validate();

			m_message = new JLabel(" ", JLabel.CENTER);
			m_message.setFont(font);
			m_message.setForeground(Color.RED);
			m_message.setSize(400,50);
			m_message.setPreferredSize(new Dimension(400,50));
			centerPanel.add( BorderLayout.SOUTH, m_message);

			contentPane.add( BorderLayout.CENTER, centerPanel);

			// --------------
			// Use a FlowLayout to center the button and give it margins.

			JPanel bottomPanel = new JPanel();

			bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

			m_buttons = new JButton[m_button_titles.length];
			for (i = 0; i < m_button_titles.length; ++i) {
				m_buttons[i] = button = new JButton(m_button_titles[i]);
				button.setFont(bold);
				tip = m_button_tips[i];
				if (tip != null) {
					button.setToolTipText(tip);
				}
				button.addActionListener(this);
				bottomPanel.add(button);
			}

			contentPane.add( BorderLayout.SOUTH, bottomPanel);

			// Resize the window to the preferred size of its components
			pack();
			setVisible(true);
		}

		public boolean ok()
		{
			return m_ret;
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object		source;
			JTextField	textfield;
			String		string, name;
			int			state, i;

			source = ev.getSource();

			state = -1;
			for (i = 0; i < m_button_titles.length; ++i) {
				if (source == m_buttons[i]) {
					state = i;
					break;
			}	}

			switch (state) {
			case BUTTON_RESET:
				for (i = 0; i < m_textfield_tags.length; ++i) {
					m_textfield_defaults[i] = m_textfield_resets[i];
				}
			case BUTTON_DEFAULT:

				for (i = 0; i < m_textfield_tags.length; ++i) {
					m_textfields[i].setText(m_textfield_defaults[i]);
				}
				break;
			case BUTTON_SET:
				for (i = 0; i < m_textfield_tags.length; ++i) {
					m_textfield_defaults[i] = m_textfields[i].getText();
				}
				m_ls.initChanged();
				break;
			case BUTTON_HELP:
				JOptionPane.showMessageDialog(m_ls.getFrame(), 	
				  "Border\n" +
				  "  Fraction of diagram to reserve for border\n" +
				  "  Must be a value between 0 and 1\n" +
				  "Gap\n" +
				  "  Fraction of diagram to reserve for whitespace"
				    , "Help", JOptionPane.OK_OPTION);
				return;
			default:

				for (i = 0; i < m_textfield_tags.length; ++i) {
					textfield = m_textfields[i];
					if (source == textfield || state == BUTTON_OK) {
						string = textfield.getText();
						string = string.trim();
						name   = m_textfield_titles[i];
						switch (i) {
						case BORDER:
						case XGAP:
						case YGAP:
							try {
								double dval;

								dval = Double.parseDouble(string);
								if (dval < 0) {
									m_message.setText(name + " may not be negative");
									return;
								}
								if (dval >= 1.0) {
									m_message.setText(name + " must be less than 1.0");
									return;
								}
							} catch (Throwable exception) {
								m_message.setText(name + " not a double precision value");
								return;
			}	}	}	}	}

			switch (state) {
			case BUTTON_OK:
				m_ret = true;
			case BUTTON_CANCEL:
				break;
			default:
				m_layoutImage.validate();
				m_layoutImage.repaint();
				return;
			}

			for (i = 0; i < m_textfield_tags.length; ++i) {
				m_textfields[i].removeActionListener(this);
			}

			setVisible(false);
			return;
		}
	}


	public MatrixLayout(LandscapeEditorCore ls) 
	{
		super(ls);
	}

	public String getName()
	{
		return "Matrix";
	}

	public String getMenuLabel() 
	{
		return "Layout in a matrix";
	} 

	public boolean isConfigurable()
	{
		return true;
	}

	public boolean configure(LandscapeEditorCore ls)
	{
		MatrixConfigure configure = new MatrixConfigure();
		boolean         ok        = configure.ok();
		configure.dispose();
		return ok;
	}

	// The doLayout method executes the Matrix algorithm
	// Assumption: All boxes selected are in the same container.

	public void doLayout1(Vector selectedBoxes, EntityInstance container) 
	{
		double			border = getBorder();
		double			xgap   = getXGap();
		double			ygap   = getYGap();

		int				size = selectedBoxes.size();
		int				rows, columns, row, col, i, gaps, actualWidth, minWidth;
		EntityInstance	e, parent, parent1;
		double			xmarginrel, ymarginrel, xrel, yrel, widthrel, heightrel, dval, gapw, gaph, last, parentWidth;
		Graphics		graphics;

		if (border < 0 || xgap < 0 || ygap < 0) {
			m_ls.error("MatrixLayout: some parameters -ve");
			return;
		}

		if (xgap >= 1.0 || ygap >= 1.0 || border >= 1.0) {
			m_ls.error("MatrixLayout: no parameter may have a value 1.0");
			return;
		}

		xmarginrel = border / 2.0;
		ymarginrel = border / 2.0;

		switch (size) {
		case 0:
			return;
		case 1:
			rows      = 1;
			break;
		default:
			rows      = (int) (Math.ceil(Math.sqrt(size)));
		}

		columns = rows;

		parent = null;
		for (i = 0; i < size; ++i) {
			e = (EntityInstance) selectedBoxes.elementAt(i);
			parent1 = e.getContainedBy();
			if (parent1 == null) {
				System.out.println("MatrixLayout: Entity " + e + " has no parent");
			} else if (parent == null) {
				parent = parent1;
			} else if (parent != parent1) {
				System.out.println("MatrixLayout: Entity " + e + " has different parent from other entities being laid out");
		}	}
		if (parent == null) {
			return;
		}

		parentWidth = (double) parent.getWidth();
		graphics    = m_ls.getGraphics();
				 
		for (; ; ) {

			if (rows == 1) {
				gaph      = 0;
				heightrel = 1.0 - border;
			} else {
				heightrel = (1.0 - border) * (1.0 - ygap) / (double) rows;
				gaph      = ((1.0 - border) - (heightrel * (double) rows))/(double) (rows-1);
			}

			if (columns == 1) {
				gapw      = 0;
				widthrel  = 1.0 - border;
				break;
			} 
			widthrel = (1.0 - border) * (1.0 - xgap) / (double) columns;
			gapw     = ((1.0 - border) - (widthrel * (double) columns))/(double) (columns-1);

			if (parentWidth <= 0) {
				// Simply layout square
				break;
			}

			for (i = 0; i < size; ++i) {
				e = (EntityInstance) selectedBoxes.elementAt(i);
				minWidth    = e.getMinFitWidth(graphics);
				actualWidth = (int) (widthrel * parentWidth);
				if (actualWidth < minWidth) {
					break;
			}	}
			if (i == size) {
				break;
			}
			--columns;
			while (rows * columns < size) {
				++rows;
		}	}

		SortVector.byString(selectedBoxes, true /* ascending */);
			
		last = 0.0;
		row = col = 0;
		for (i = 0; i < size; ++i) {
			e = (EntityInstance) selectedBoxes.elementAt(i);
			xrel = xmarginrel;
			if (col != 0) {
				xrel += ((double) col) * (widthrel + gapw);
			}
			last = xrel;
			yrel = ymarginrel;
			if (row != 0) {
				yrel += ((double) row)*(heightrel + gaph);
			}
			e.updateRelLocal(xrel, yrel, widthrel, heightrel);
			if (++col == columns) {
				++row;
				col = 0;
		}	}

//		System.out.println("Matrix done");

	} // doLayout


  // The doLayout method executes the Coffman-Graham Layer Assignment
  // algorithm and the Sugiyama algorithm on the boxes selected.
  // Assumption: All boxes selected are in the same container.

	public String doLayout(Diagram dg) 
	{
		EntityInstance parent;

		// get user's selection of boxes to be laid out

		m_ls.setLayouter(this);

		Vector selectedBoxes = dg.getGroup();
		if (selectedBoxes == null) {
			  Util.beep();
			  return "No group selected";
		}

		String msg = allInDiagram(selectedBoxes);
		if (msg != null) {
			return msg;
		}

		parent = parentOfSet(selectedBoxes);
		if (parent == null) {
			return	"Matrix layouter requires that all things laid out share same parent";
		}
		m_ls.setCursor(Cursor.WAIT_CURSOR);

		dg.beginUndoRedo("Matrix layout");
		doLayout1(selectedBoxes, parent);
		dg.endUndoRedo();

		m_ls.setCursor(Cursor.DEFAULT_CURSOR);

		return "Graph redrawn using Matrix Layout";
	} // doLayout

	public void processKeyEvent(int key, int modifiers, Object object) 
	{
		Diagram	dg;
		String	rmsg;

/*
		if (!configure(m_ls)) {
			return;
		}
*/
		dg = m_ls.getDiagram();
		if (dg != null) {
			rmsg = doLayout(dg);
			m_ls.doFeedback(rmsg);
			dg.rescaleDiagram();
			m_ls.repaintDg();
	}	}
} 





